@@ -4,13 +4,19 @@ from collections import OrderedDict
 
 import tensorflow.keras.backend as K
 from tensorflow.keras.models import Model
+from tensorflow.keras.layers import Layer
 
 
-def n_(node, output_format_):
+def n_(node, output_format_, nested=False):
     node_name = str(node.name)
     if output_format_ == 'simple':
         if '/' in node_name:
-            return node_name.split('/')[0]
+            # This ensures that subnodes get properly named.
+            tokens = node_name.split('/')
+            if nested:
+                return '/'.join(tokens[:-1])
+            else:
+                return tokens[0]
         elif ':' in node_name:
             return node_name.split(':')[0]
         else:
@@ -60,12 +66,11 @@ def get_gradients_of_trainable_weights(model, x, y):
     :param y: outputs for which gradients are sought
     :return: dict mapping layers to corresponding gradients (filter_h, filter_w, in_channels, out_channels)
     """
-    nodes = model.trainable_weights
-    nodes_names = [w.name for w in nodes]
-    return _get_gradients(model, x, y, nodes, nodes_names)
+    nodes = OrderedDict([(n.name, n) for n in model.trainable_weights])
+    return _get_gradients(model, x, y, nodes)
 
 
-def get_gradients_of_activations(model, x, y, layer_name=None, output_format='simple'):
+def get_gradients_of_activations(model, x, y, layer_names=None, output_format='simple', nested=False):
     """
     Get gradients of the outputs of the activation functions, regarding the loss.
     Intuitively, it shows how your activation maps change over a tiny modification of the loss.
@@ -73,61 +78,122 @@ def get_gradients_of_activations(model, x, y, layer_name=None, output_format='si
     'mobilenet_v2', 'mobilenetv2'].
     :param x: Model input (Numpy array). In the case of multi-inputs, x should be of type List.
     :param y: Model target (Numpy array). In the case of multi-inputs, y should be of type List.
-    :param layer_name: (optional) Name of a layer for which activations should be returned only. It is useful in
-    very big networks when it is computationally expensive to evaluate all the layers/nodes.
+    :param layer_names: (optional) Single name of a layer or list of layer names for which activations should be
+    returned. It is useful in very big networks when it is computationally expensive to evaluate all the layers/nodes.
     :param output_format: Change the output dictionary key of the function.
     - 'simple': output key will match the names of the Keras layers. For example Dense(1, name='d1') will
     return {'d1': ...}.
     - 'full': output key will match the full name of the output layer name. In the example above, it will
     return {'d1/BiasAdd:0': ...}.
     - 'numbered': output key will be an index range, based on the order of definition of each layer within the model.
-    :return: Dict {layer_name (specified by output_format) -> activation of the layer output/node (Numpy array)}.
+    - 'nested': If specified, will move recursively through the model definition to retrieve nested layers.
+                Recursion ends at leaf layers of the model tree or at layers with their name specified in layer_names.
+
+                E.g., a model with the following structure
+
+                -layer1
+                    -conv1
+                    ...
+                    -fc1
+                -layer2
+                    -fc2
+
+                ... yields a dictionary with keys 'layer1/conv1', ..., 'layer1/fc1', 'layer2/fc2'.
+                If layer_names = ['layer2/fc2'] is specified, the dictionary will only hold one key 'layer2/fc2'.
+
+                The layer names are generated by joining all layers from top level to leaf level with the separator '/'.
+    :return: Dict {layer_names (specified by output_format) -> activation of the layer output/node (Numpy array)}.
     """
-    nodes = [layer.output for layer in model.layers if layer.name == layer_name or layer_name is None]
-    return _get_gradients(model, x, y, nodes, output_format)
+    nodes = _get_nodes(model, output_format, nested=nested, layer_names=layer_names)
+    return _get_gradients(model, x, y, nodes)
 
 
-def _get_gradients(model, x, y, nodes, output_format):
+def _get_gradients(model, x, y, nodes):
     if model.optimizer is None:
         raise Exception('Please compile the model first. The loss function is required to compute the gradients.')
+
+    nodes_names = nodes.keys()
+    nodes_values = nodes.values()
+
     try:
         if not hasattr(model, 'total_loss'):
             raise Exception('Disable TF eager mode to use get_gradients.\n'
                             'Add this command at the beginning of your script:\n'
                             'tf.compat.v1.disable_eager_execution()')
-        grads = model.optimizer.get_gradients(model.total_loss, nodes)
+        grads = model.optimizer.get_gradients(model.total_loss, nodes_values)
     except ValueError as e:
         if 'differentiable' in str(e):
             # Probably one of the gradients operations is not differentiable...
             grads = []
             differentiable_nodes = []
-            for n in nodes:
+            for n in nodes_values:
                 try:
                     grads.extend(model.optimizer.get_gradients(model.total_loss, n))
                     differentiable_nodes.append(n)
                 except ValueError:
                     pass
-            nodes = differentiable_nodes
+            nodes_values = differentiable_nodes
         else:
             raise e
 
     gradients_values = _evaluate(model, grads, x, y)
-    nodes_names = [n_(n, output_format) for n in nodes]
-    if len(set(nodes_names)) != len(nodes):  # collision detected.
-        nodes_names = [n_(n, 'full') for n in nodes]
+
     return OrderedDict(zip(nodes_names, gradients_values))
 
 
-def get_activations(model, x, layer_name=None, nodes_to_evaluate=None,
-                    output_format='simple', auto_compile=True):
+def _get_nodes(module, output_format, nested=False, layer_names=[]):
+    is_model_or_layer = isinstance(module, Model) or isinstance(module, Layer)
+    has_layers = hasattr(module, '_layers') and bool(module._layers)
+    assert is_model_or_layer, 'Not a model or layer!'
+
+    module_name = n_(module.output, output_format_=output_format, nested=nested)
+
+    if has_layers:
+        node_dict = OrderedDict()
+        # print('Layers:', module._layers)
+        for m in module._layers:
+            key = n_(m.output, output_format_=output_format, nested=nested)
+            if nested:
+                nodes = _get_nodes(m, output_format,
+                                   nested=nested,
+                                   layer_names=layer_names)
+            else:
+                if bool(layer_names) and key in layer_names:
+                    nodes = OrderedDict([(key, m.output)])
+                elif not bool(layer_names):
+                    nodes = OrderedDict([(key, m.output)])
+                else:
+                    nodes = OrderedDict()
+            node_dict.update(nodes)
+        return node_dict
+
+    elif bool(layer_names) and module_name in layer_names:
+        print("1", module_name, module)
+        return OrderedDict({module_name: module.output})
+
+    elif not bool(layer_names):
+        print("2", module_name, module)
+        return OrderedDict({module_name: module.output})
+
+    else:
+        print("3", module_name, module)
+        return OrderedDict()
+
+
+def _is_input(layer, key):
+    return 'input' in key
+
+
+def get_activations(model, x, layer_names=None, nodes_to_evaluate=None,
+                    output_format='simple', nested=False, auto_compile=True):
     """
     Fetch activations (nodes/layers outputs as Numpy arrays) for a Keras model and an input X.
     By default, all the activations for all the layers are returned.
     :param model: Keras compiled model or one of ['vgg16', 'vgg19', 'inception_v3', 'inception_resnet_v2',
     'mobilenet_v2', 'mobilenetv2', ...].
     :param x: Model input (Numpy array). In the case of multi-inputs, x should be of type List.
-    :param layer_name: (optional) Name of a layer for which activations should be returned only. It is useful in
-    very big networks when it is computationally expensive to evaluate all the layers/nodes.
+    :param layer_names: (optional) Single name of a layer or list of layer names for which activations should be
+    returned. It is useful in very big networks when it is computationally expensive to evaluate all the layers/nodes.
     :param nodes_to_evaluate: (optional) List of Keras nodes to be evaluated. Useful when the nodes are not
     in model.layers.
     :param output_format: Change the output dictionary key of the function.
@@ -136,50 +202,72 @@ def get_activations(model, x, layer_name=None, nodes_to_evaluate=None,
     - 'full': output key will match the full name of the output layer name. In the example above, it will
     return {'d1/BiasAdd:0': ...}.
     - 'numbered': output key will be an index range, based on the order of definition of each layer within the model.
+    - 'nested': If specified, will move recursively through the model definition to retrieve nested layers.
+                Recursion ends at leaf layers of the model tree or at layers with their name specified in layer_names.
+
+                E.g., a model with the following structure
+
+                -layer1
+                    -conv1
+                    ...
+                    -fc1
+                -layer2
+                    -fc2
+
+                ... yields a dictionary with keys 'layer1/conv1', ..., 'layer1/fc1', 'layer2/fc2'.
+                If layer_names = ['layer2/fc2'] is specified, the dictionary will only hold one key 'layer2/fc2'.
+
+                The layer names are generated by joining all layers from top level to leaf level with the separator '/'.
     :param auto_compile: If set to True, will auto-compile the model if needed.
     :return: Dict {layer_name (specified by output_format) -> activation of the layer output/node (Numpy array)}.
     """
+    layer_names = [layer_names] if isinstance(layer_names, str) else layer_names
+    # print('Layer names:', layer_names)
     if nodes_to_evaluate is None:
-        nodes = [layer.output for layer in model.layers if layer.name == layer_name or layer_name is None]
+        nodes = _get_nodes(model, output_format, layer_names=layer_names, nested=nested)
     else:
-        if layer_name is not None:
+        if layer_names is not None:
             raise ValueError('Do not specify a [layer_name] with [nodes_to_evaluate]. It will not be used.')
-        nodes = nodes_to_evaluate
+        nodes = OrderedDict([(n_(node, 'full'), node) for node in nodes_to_evaluate])
 
     if len(nodes) == 0:
-        if layer_name is not None:
+        if layer_names is not None:
             network_layers = ', '.join([layer.name for layer in model.layers])
             raise KeyError('Could not find a layer with name: [{}]. '
-                           'Network layers are [{}]'.format(layer_name, network_layers))
+                           'Network layers are [{}]'.format(', '.join(layer_names), network_layers))
         else:
             raise ValueError('Nodes list is empty. Or maybe the model is empty.')
 
     # The placeholders are processed later (Inputs node in Keras). Due to a small bug in tensorflow.
-    input_layer_outputs, layer_outputs = [], []
-    for node in nodes:
+    input_layer_outputs = []
+    layer_outputs = OrderedDict()
+
+    for key, node in nodes.items():
         if node.op.type != 'Placeholder':  # no inputs please.
-            layer_outputs.append(node)
-    if nodes_to_evaluate is None or (layer_name is not None) and \
-            any([n.name.startswith(layer_name) for n in model.inputs]):
+            layer_outputs.update({key: node})
+    if nodes_to_evaluate is None or (layer_names is not None) and \
+            any([n.name in layer_names for n in model.inputs]):
         input_layer_outputs = list(model.inputs)
+
     if len(layer_outputs) > 0:
-        activations = _evaluate(model, layer_outputs, x, y=None, auto_compile=auto_compile)
+        activations = _evaluate(model, layer_outputs.values(), x, y=None, auto_compile=auto_compile)
     else:
         activations = {}
 
     def craft_output(output_format_):
-        activations_dict = OrderedDict(zip([n_(output, output_format_) for output in layer_outputs], activations))
         inputs = [x] if not isinstance(x, list) else x
         activations_inputs_dict = OrderedDict(zip([n_(output, output_format_) for output in input_layer_outputs], inputs))
+        activations_dict = OrderedDict(zip(layer_outputs.keys(), activations))
         result_ = activations_inputs_dict.copy()
         result_.update(activations_dict)
+
         if output_format_ == 'numbered':
             result_ = OrderedDict([(i, v) for i, (k, v) in enumerate(result_.items())])
         return result_
 
     result = craft_output(output_format)
-    if layer_name is not None:  # extra check.
-        result = {k: v for k, v in result.items() if k == layer_name}
+    if layer_names is not None:  # extra check.
+        result = {k: v for k, v in result.items() if k in layer_names}
     if nodes_to_evaluate is not None and len(result) != len(nodes_to_evaluate):
         result = craft_output(output_format_='full')  # collision detected in the keys.
 
